package org.bitbucket.bradleysmithllc.etlunit;

import org.bitbucket.bradleysmithllc.etlunit.parser.ETLTestAnnotation;
import org.bitbucket.bradleysmithllc.etlunit.parser.ETLTestClass;
import org.bitbucket.bradleysmithllc.etlunit.parser.ETLTestMethod;
import org.bitbucket.bradleysmithllc.etlunit.parser.ETLTestOperation;
import org.bitbucket.bradleysmithllc.etlunit.regexp.SuiteNameExpression;
import org.bitbucket.bradleysmithllc.etlunit.regexp.SuiteNamesExpression;

import java.util.ArrayList;
import java.util.List;

public class SuiteClassDirectorImpl extends NullClassDirector {
	public static enum required_type {
		include,
		exclude,
		optional
	}

	public static final class SuiteRef {
		enum accept {yes, no , maybe}

		private final required_type type;
		private final String suiteName;

		public SuiteRef(SuiteNameExpression sn) {
			suiteName = sn.getSuiteName().toLowerCase();

			String ot = sn.hasOptionType() ? sn.getOptionType() : null;

			if (ot == null || ot.equals("+")) {
				type = required_type.include;
			} else if (ot.equals("?")) {
				type = required_type.optional;
			}	else {
				type = required_type.exclude;
			}
		}

		public SuiteRef(String name, required_type type) {
			suiteName = name.toLowerCase();

			this.type = type != null ? type : required_type.include;
		}

		public accept accept(List<ETLTestAnnotation> annoList) {
			// check against this name
			for (ETLTestAnnotation anno : annoList) {
				String name = anno.getValue().query("name").getValueAsString();

				if (suiteName.equals(name.toLowerCase())) {
					// decide if this is good or bad
					switch (type) {
						case optional:
							// match means include
						case include:
							// match means good
							return accept.yes;
						case exclude:
							// match means bad
							return accept.no;
					}
				}
			}

			// no match can go either way
			switch (type) {
				case include:
					// no match means bad
					return accept.no;
				case optional:
					// match means maybe include.  An all maybe response means exclude
					return accept.maybe;
				case exclude:
					// no match means good
				default:
					return accept.yes;
			}
		}

		public required_type getType() {
			return type;
		}

		public String getSuiteName() {
			return suiteName;
		}
	}

	private final List<SuiteRef> suiteRefs = new ArrayList<SuiteRef>();

	public SuiteClassDirectorImpl(String text) {
		SuiteNamesExpression sne = new SuiteNamesExpression(text);

		if (sne.hasNext()) {
			SuiteNameExpression sn = sne.getSuiteName();

			while (sn.hasNext()) {
				suiteRefs.add(new SuiteRef(sn));
			}
		}
	}

	public SuiteClassDirectorImpl() {
	}

	public void addSuite(String name, required_type type)
	{
		suiteRefs.add(new SuiteRef(name, type));
	}

	@Override
	public response_code accept(ETLTestClass cl) {
		List<ETLTestAnnotation> annoList = cl.getAnnotations("@JoinSuite");

		if (annoList.size() == 0 || suiteRefs.size() == 0) {
			return response_code.reject;
		}

		SuiteRef.accept bestAccept = SuiteRef.accept.no;

		for (SuiteRef ref : suiteRefs) {
			SuiteRef.accept accept = ref.accept(annoList);
			switch (accept) {

				case yes:
					bestAccept = SuiteRef.accept.yes;
					break;
				case no:
					return response_code.reject;
				case maybe:
					if (bestAccept != SuiteRef.accept.yes)
					{
						bestAccept = SuiteRef.accept.maybe;
					}

					break;
			}
		}

		return bestAccept == SuiteRef.accept.yes ? response_code.accept : response_code.reject;
	}

	@Override
	public response_code accept(ETLTestMethod mt) {
		return response_code.accept;
	}

	@Override
	public response_code accept(ETLTestOperation op) {
		return response_code.accept;
	}

	public List<SuiteRef> getSuiteRefs()
	{
		return suiteRefs;
	}
}
